@@ -27,7 +27,10 @@
 #include <gconf/gconf-client.h>
 #include <osso-ic-dbus.h>
 #include <osso-ic-gconf.h>
-#include <osso-ic.h>
+
+#ifndef OSSO_IAP_ANY
+#define OSSO_IAP_ANY	"[ANY]"
+#endif
 
 #include "conicevent.h"
 #include "conicevent-private.h"
@@ -40,6 +43,10 @@
 #undef G_LOG_DOMAIN
 #define G_LOG_DOMAIN "ConIc"
 
+#define ICD_DBUS_MATCH		"type='signal'," \
+				"interface='" ICD_DBUS_INTERFACE "'," \
+				"path='" ICD_DBUS_PATH "'"
+
 typedef enum {
 	CONNECTING,
 	CONNECTED,
@@ -47,7 +54,8 @@ typedef enum {
 	DISCONNECTED,
 } ConIcConnectionInternalStatus;
 
-GSList *objects = NULL;
+static GSList *objects = NULL;
+static DBusConnection *dbus_connection = NULL;
 
 struct _ConIcConnectionClass
 {
@@ -69,6 +77,7 @@ struct _ConIcConnection
 		*autoconfig_url;
 	gint http_port, ftp_port, https_port, socks_port, rtsp_port;
 	GSList *ignore_hosts;
+	GSList *pending_calls;
 };
 
 static GObjectClass *parent_class = NULL;
@@ -215,6 +224,12 @@ static void con_ic_connection_finalize(GObject *obj)
 {
 	ConIcConnection *connection = CON_IC_CONNECTION(obj);
 
+	g_slist_foreach(connection->pending_calls, 
+		        (GFunc)dbus_pending_call_cancel, NULL);
+	g_slist_foreach(connection->pending_calls, 
+		        (GFunc)dbus_pending_call_unref, NULL);
+	g_slist_free(connection->pending_calls);
+	
 	g_free(connection->iap_id);
 	g_free(connection->bearer);
 	g_free(connection->http_host);
@@ -225,6 +240,15 @@ static void con_ic_connection_finalize(GObject *obj)
 	g_free(connection->autoconfig_url);
 
 	objects = g_slist_remove(objects, connection);
+	if (objects == NULL) {
+		/* If all objects were deleted, unref DBUS stuff */
+		dbus_connection_unregister_object_path(dbus_connection,
+						       ICD_DBUS_PATH);
+		dbus_bus_remove_match(dbus_connection, ICD_DBUS_MATCH, NULL);
+		
+		dbus_connection_unref(dbus_connection);
+		dbus_connection = NULL;
+	}
 }
 
 static void con_ic_connection_state_cb(DBusPendingCall *pending,
@@ -238,6 +262,12 @@ static void con_ic_connection_state_cb(DBusPendingCall *pending,
 
 	dbus_error_init(&error);
 
+	if (g_slist_find(connection->pending_calls, pending)) {
+		connection->pending_calls = g_slist_remove(
+			connection->pending_calls, pending);
+		dbus_pending_call_unref(pending);
+	}
+
 	reply = dbus_pending_call_steal_reply(pending);
 
 	if (reply == NULL) {
@@ -298,10 +328,13 @@ static gboolean con_ic_connection_enable_automatic_events(ConIcConnection *conne
 	if (!dbus_pending_call_set_notify(pending,
 					  con_ic_connection_state_cb,
 					  connection, NULL)) {
+		dbus_pending_call_unref(pending);
 		ret = FALSE;
 		goto exit;
 	}
 	
+	connection->pending_calls = g_slist_insert(
+		connection->pending_calls, pending, 0);
 	ret = TRUE;
 	
 exit:
@@ -357,7 +390,7 @@ static void con_ic_connection_send_event(ConIcConnection *connection,
 					 ConIcConnectionStatus status)
 {
 	g_debug("con_ic_connection_send_event(%p, %s, %s, %i)",
-		connection, id, bearer, status);
+		(void *)connection, id, bearer, status);
 	
 	ConIcConnectionEvent *event = con_ic_connection_event_new();
 
@@ -368,6 +401,7 @@ static void con_ic_connection_send_event(ConIcConnection *connection,
 	g_signal_emit(connection,
 		      CON_IC_CONNECTION_GET_CLASS(connection)->connection_signal_id,
 		      0, event);
+	g_object_unref(G_OBJECT(event));
 }
 
 static DBusHandlerResult con_ic_connection_handle_icd_message(DBusConnection *c,
@@ -376,12 +410,10 @@ static DBusHandlerResult con_ic_connection_handle_icd_message(DBusConnection *c,
 {
 	char *iap_id, *bearer, *state;
 	ConIcConnectionStatus status;
-	DBusHandlerResult ret;
-	GSList *e;
+	GSList *e, *copy = NULL;
 	
 	if (!dbus_message_is_signal(msg, ICD_DBUS_INTERFACE, 
 				    ICD_STATUS_CHANGED_SIG)) {
-		ret = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		goto exit;
 	}
 
@@ -390,7 +422,6 @@ static DBusHandlerResult con_ic_connection_handle_icd_message(DBusConnection *c,
 				   DBUS_TYPE_STRING, &bearer,
 				   DBUS_TYPE_STRING, &state,
 				   DBUS_TYPE_INVALID)) {
-		ret = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		goto exit;
 	}
 
@@ -399,12 +430,18 @@ static DBusHandlerResult con_ic_connection_handle_icd_message(DBusConnection *c,
 	} else if (strcmp(state, "CONNECTED") == 0) {
 		status = CON_IC_STATUS_CONNECTED;
 	} else {
-		ret = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 		goto exit;
 	}
 	
+	/* ref all objects in the list so they don't get destroyed, and
+	 * take a copy so we remember which objects have been reffed;
+	 * con_ic_connection_send_event might lead to unref on any of the
+	 * objects in the list or addition of objects */
+	copy = g_slist_copy(objects);
+	g_slist_foreach(copy, (GFunc)g_object_ref, NULL);
+	
 	/* is this process connected already? */
-	for (e = objects; e != NULL; e = g_slist_next(e)) {
+	for (e = copy; e != NULL; e = g_slist_next(e)) {
 		ConIcConnection *connection = e->data;
 
 		/* if automatic events are enabled, application has to track
@@ -440,16 +477,15 @@ static DBusHandlerResult con_ic_connection_handle_icd_message(DBusConnection *c,
 		}
 	}
 	
-	ret = DBUS_HANDLER_RESULT_HANDLED;
-
 exit:
-	return ret;
+	g_slist_foreach(copy, (GFunc)g_object_unref, NULL);
+	g_slist_free(copy);
+
+	return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
 }
 
 static DBusConnection *con_ic_connection_get_dbus_connection(ConIcConnection *connection)
 {
-	static DBusConnection *dbus_connection = NULL;
-	
 	static struct DBusObjectPathVTable icd_vtable = {
 		.message_function = &con_ic_connection_handle_icd_message
 	};
@@ -462,18 +498,14 @@ static DBusConnection *con_ic_connection_get_dbus_connection(ConIcConnection *co
 			goto exit;
 		}
 
-
 		dbus_bus_add_match(dbus_connection,
-				   "type='signal',"
-				   "interface='" ICD_DBUS_INTERFACE "',"
-				   "path='" ICD_DBUS_PATH "'",
+				   ICD_DBUS_MATCH,
 				   NULL);
 		/* FIXME: pass a list as user_data */
 		if (!dbus_connection_register_object_path(dbus_connection,
 							  ICD_DBUS_PATH,
 							  &icd_vtable,
 							  NULL)) {
-			dbus_connection_disconnect(dbus_connection);
 			dbus_connection_unref(dbus_connection);
 			dbus_connection = NULL;
 		}
@@ -496,6 +528,12 @@ static void con_ic_connection_connected_cb(DBusPendingCall *pending,
 	
 	dbus_error_init(&error);
 
+	if (g_slist_find(connection->pending_calls, pending)) {
+		connection->pending_calls = g_slist_remove(
+			connection->pending_calls, pending);
+		dbus_pending_call_unref(pending);
+	}
+
 	reply = dbus_pending_call_steal_reply(pending);
 
 	if (reply == NULL) {
@@ -554,6 +592,12 @@ static void con_ic_connection_disconnected_cb(DBusPendingCall *pending,
 
 	dbus_error_init(&error);
 
+	if (g_slist_find(connection->pending_calls, pending)) {
+		connection->pending_calls = g_slist_remove(
+			connection->pending_calls, pending);
+		dbus_pending_call_unref(pending);
+	}
+
 	reply = dbus_pending_call_steal_reply(pending);
 
 	if (reply == NULL) {
@@ -607,6 +651,12 @@ static void con_ic_connection_statistics_cb(DBusPendingCall *pending,
 
 	dbus_error_init(&error);
 
+	if (g_slist_find(connection->pending_calls, pending)) {
+		connection->pending_calls = g_slist_remove(
+			connection->pending_calls, pending);
+		dbus_pending_call_unref(pending);
+	}
+
 	reply = dbus_pending_call_steal_reply(pending);
 
 	if (reply == NULL) {
@@ -676,6 +726,8 @@ gboolean con_ic_connection_connect_by_id(ConIcConnection *connection,
 	GSList *e;
 	
 	/* is this process connected already? */
+	/* No need to copy objects here as no connection nor objects are
+	   accessed after send_event() */
 	for (e = objects; e != NULL; e = g_slist_next(e)) {
 		ConIcConnection *c = e->data;
 		if (c == connection) {
@@ -734,10 +786,13 @@ gboolean con_ic_connection_connect_by_id(ConIcConnection *connection,
 	if (!dbus_pending_call_set_notify(pending,
 					  con_ic_connection_connected_cb,
 					  connection, NULL)) {
+		dbus_pending_call_unref(pending);
 		ret = FALSE;
 		goto exit;
 	}
 
+	connection->pending_calls = g_slist_insert(
+		connection->pending_calls, pending, 0);
 	connection->internal_status = CONNECTING;
 	ret = TRUE;
 	
@@ -745,9 +800,6 @@ exit:
 	if (msg)
 		dbus_message_unref(msg);
 
-	if (pending)
-		dbus_pending_call_unref(pending);
-
 	return ret;
 }
 
@@ -822,10 +874,13 @@ gboolean con_ic_connection_disconnect_by_id(ConIcConnection *connection,
 	if (!dbus_pending_call_set_notify(pending,
 					  con_ic_connection_disconnected_cb,
 					  connection, NULL)) {
+		dbus_pending_call_unref(pending);
 		ret = FALSE;
 		goto exit;
 	}
 
+	connection->pending_calls = g_slist_insert(
+		connection->pending_calls, pending, 0);
 	ret = TRUE;
 	connection->internal_status = DISCONNECTING;
 	
@@ -833,9 +888,6 @@ exit:
 	if (msg)
 		dbus_message_unref(msg);
 
-	if (pending)
-		dbus_pending_call_unref(pending);
-
 	return ret;
 }
 
@@ -874,19 +926,19 @@ gboolean con_ic_connection_statistics(ConIcConnection *connection,
 	if (!dbus_pending_call_set_notify(pending,
 					  con_ic_connection_statistics_cb,
 					  connection, NULL)) {
+		dbus_pending_call_unref(pending);
 		ret = FALSE;
 		goto exit;
 	}
 
+	connection->pending_calls = g_slist_insert(
+		connection->pending_calls, pending, 0);
 	ret = TRUE;
 	
 exit:
 	if (msg)
 		dbus_message_unref(msg);
 
-	if (pending)
-		dbus_pending_call_unref(pending);
-
 	return ret;
 }
 
